{-# OPTIONS --safe --without-K #-}

open import Data.Nat.Base using (ℕ)

module Regex.Properties.Sub (α : ℕ) where

open import Regex α
open import Regex.Properties.Core α

open import Algebra.Definitions _⊑_
open import Data.List.Base using ([]; _∷_)
open import Data.Sum.Base using (inj₁; inj₂)
open import Relation.Binary.Definitions
open import Relation.Nullary.Decidable.Core using (no; yes)
open import Relation.Nullary.Negation.Core using (contradiction)

------------------------------------------------------------------------
-- General properties

refl : Reflexive _⊑_
refl as p = p

trans : Transitive _⊑_
trans x y as p = y as (x as p)

------------------------------------------------------------------------
-- Algebraic properties

-- Properties of _+_

+-cong : Congruent₂ _+_
+-cong x⊑y u⊑v as p with split as p
... | inj₁ p′ = join₁ as (x⊑y as p′)
... | inj₂ p′ = join₂ as (u⊑v as p′)

+-congˡ : LeftCongruent _+_
+-congˡ x⊑y = +-cong refl x⊑y

+-congʳ : RightCongruent _+_
+-congʳ x⊑y = +-cong x⊑y refl

+-assoc : Associative _+_
+-assoc r₁ r₂ r₃ [] (+ₗ (+ₗ p)) = +ₗ p
+-assoc r₁ r₂ r₃ [] (+ₗ (+ᵣ p)) = +ᵣ (+ₗ p)
+-assoc r₁ r₂ r₃ [] (+ᵣ p) = +ᵣ (+ᵣ p)
+-assoc r₁ r₂ r₃ (a ∷ as) p = +-assoc _ _ _ as p

+-comm : Commutative _+_
+-comm r₁ r₂ [] (+ₗ p) = +ᵣ p
+-comm r₁ r₂ [] (+ᵣ p) = +ₗ p
+-comm r₁ r₂ (a ∷ as) p = +-comm _ _ as p

+-idem : Idempotent _+_
+-idem r [] (+ₗ p) = p
+-idem r [] (+ᵣ p) = p
+-idem r (a ∷ as) p = +-idem _ as p

+-identityˡ : LeftIdentity Λ _+_
+-identityˡ r [] (+ᵣ p) = p
+-identityˡ r (a ∷ as) p = +-identityˡ _ as p

+-identityʳ : RightIdentity Λ _+_
+-identityʳ r [] (+ₗ p) = p
+-identityʳ r (a ∷ as) p = +-identityʳ _ as p

-- Properties of _⨟_

⨟-cong : Congruent₂ _⨟_
⨟-cong x⊑y u⊑v [] (p₁ ⨟ p₂) = x⊑y [] p₁ ⨟ u⊑v [] p₂
⨟-cong {x = x} {y = y} x⊑y u⊑v (a ∷ as) p with A? x | A? y
... | no ¬Ax | no ¬Ay = ⨟-cong (λ as′ → x⊑y (a ∷ as′)) u⊑v as p
... | no ¬Ax | yes Ay = join₂ as (⨟-cong (λ as′ → x⊑y (a ∷ as′)) u⊑v as p)
... | yes Ax | no ¬Ay = contradiction (x⊑y [] Ax) ¬Ay
... | yes Ax | yes Ay with split as p
... | inj₁ p′ = join₁ as (u⊑v (a ∷ as) p′)
... | inj₂ p′ = join₂ as (⨟-cong (λ as′ → x⊑y (a ∷ as′)) u⊑v as p′)

⨟-congʳ : RightCongruent _⨟_
⨟-congʳ x⊑y = ⨟-cong x⊑y refl

⨟-distribˡ-+ : _⨟_ DistributesOverˡ _+_
⨟-distribˡ-+ r₁ r₂ r₃ [] (p₁ ⨟ +ₗ p₂) = +ₗ (p₁ ⨟ p₂)
⨟-distribˡ-+ r₁ r₂ r₃ [] (p₁ ⨟ +ᵣ p₂) = +ᵣ (p₁ ⨟ p₂)
⨟-distribˡ-+ r₁ r₂ r₃ (a ∷ as) p with A? r₁
... | no ¬Ar₁ = ⨟-distribˡ-+ _ _ _ as p
... | yes Ar₁ with split as p
... | inj₁ p′ with split as p′
... | inj₁ p″ = join₁ as (join₁ as p″)
... | inj₂ p″ = join₂ as (join₁ as p″)
⨟-distribˡ-+ r₁ r₂ r₃ (a ∷ as) p | yes Ar₁ | inj₂ p′ with split as (⨟-distribˡ-+ _ _ _ as p′)
... | inj₁ p″ = join₁ as (join₂ as p″)
... | inj₂ p″ = join₂ as (join₂ as p″)

⨟-distribʳ-+ : _⨟_ DistributesOverʳ _+_
⨟-distribʳ-+ r₁ r₂ r₃ [] (+ₗ p₁ ⨟ p₂) = +ₗ (p₁ ⨟ p₂)
⨟-distribʳ-+ r₁ r₂ r₃ [] (+ᵣ p₁ ⨟ p₂) = +ᵣ (p₁ ⨟ p₂)
⨟-distribʳ-+ r₁ r₂ r₃ (a ∷ as) p with A? r₂ | A? r₃
... | no ¬Ar₂ | no ¬Ar₃ = ⨟-distribʳ-+ r₁ (δ r₂ a) (δ r₃ a) as p
... | no ¬Ar₂ | yes Ar₃ with split as p
... | inj₁ p′ = join₂ as (join₁ as p′)
... | inj₂ p′ with split as (⨟-distribʳ-+ r₁ _ _ as p′)
... | inj₁ p″ = join₁ as p″
... | inj₂ p″ = join₂ as (join₂ as p″)
⨟-distribʳ-+ r₁ r₂ r₃ (a ∷ as) p | yes Ar₂ | no ¬Ar₃ with split as p
... | inj₁ p′ = join₁ as (join₁ as p′)
... | inj₂ p′ with split as (⨟-distribʳ-+ r₁ (δ r₂ a) (δ r₃ a) as p′)
... | inj₁ p″ = join₁ as (join₂ as p″)
... | inj₂ p″ = join₂ as p″
⨟-distribʳ-+ r₁ r₂ r₃ (a ∷ as) p | yes Ar₂ | yes Ar₃ with split as p
... | inj₁ p′ = join₁ as (join₁ as p′)
... | inj₂ p′ with split as (⨟-distribʳ-+ _ _ _ as p′)
... | inj₁ p″ = join₁ as (join₂ as p″)
... | inj₂ p″ = join₂ as (join₂ as p″)

⨟-assoc : Associative _⨟_
⨟-assoc r₁ r₂ r₃ [] ((p₁ ⨟ p₂) ⨟ p₃) = p₁ ⨟ (p₂ ⨟ p₃)
⨟-assoc r₁ r₂ r₃ (a ∷ as) p with A? r₁
... | no ¬Ar₁ = ⨟-assoc (δ r₁ a) r₂ r₃ as p
... | yes Ar₁ with A? r₂
... | no ¬Ar₂ with split as (⨟-distribʳ-+ _ _ _ as p)
... | inj₁ p′ = join₁ as p′
... | inj₂ p′ = join₂ as (⨟-assoc _ _ _ as p′)
⨟-assoc r₁ r₂ r₃ (a ∷ as) p | yes Ar₁ | yes Ar₂ with split as p
... | inj₁ p′ = join₁ as (join₁ as p′)
... | inj₂ p′ with split as (⨟-distribʳ-+ _ _ _ as p′)
... | inj₁ p″ = join₁ as (join₂ as p″)
... | inj₂ p″ = join₂ as (⨟-assoc _ _ _ as p″)

⨟-zeroˡ : LeftZero Λ _⨟_
⨟-zeroˡ r [] (() ⨟ _)
⨟-zeroˡ r (a ∷ as) p = ⨟-zeroˡ r as p

⨟-zeroʳ : RightZero Λ _⨟_
⨟-zeroʳ r [] (p₁ ⨟ p₂) = p₂
⨟-zeroʳ r (a ∷ as) p with A? r
... | no ¬Ar = ⨟-zeroʳ _ as p
... | yes Ar with split as p
... | inj₁ p′ = p′
... | inj₂ p′ = ⨟-zeroʳ _ as p′

⨟-identityˡ : LeftIdentity ε _⨟_
⨟-identityˡ r [] (star ⨟ p) = p
⨟-identityˡ r (a ∷ as) p with split as p
... | inj₁ p′ = p′
... | inj₂ p′ = contradiction (⨟-zeroˡ _ as (⨟-assoc _ _ _ as p′)) (end as)

⨟-identityʳ : RightIdentity ε _⨟_
⨟-identityʳ r [] (p ⨟ star) = p
⨟-identityʳ r (a ∷ as) p with A? r
... | no ¬Ar = ⨟-identityʳ _ as p
... | yes Ar with split as p
... | inj₁ p′ = contradiction (⨟-zeroˡ _ as p′) (end as)
... | inj₂ p′ = ⨟-identityʳ _ as p′
