{-# OPTIONS --safe --without-K #-}

open import Data.Nat.Base using (ℕ)

module Regex.Properties.Super (α : ℕ) where

open import Regex α
open import Regex.Properties.Core α

open import Algebra.Definitions _⊒_
open import Data.List.Base using ([]; _∷_)
open import Data.Sum.Base using (inj₁; inj₂)
open import Relation.Binary.Definitions
open import Relation.Nullary.Decidable.Core using (no; yes)
open import Relation.Nullary.Negation.Core using (contradiction)

------------------------------------------------------------------------
-- General properties

refl : Reflexive _⊒_
refl as p = p

trans : Transitive _⊒_
trans x y as p = x as (y as p)

------------------------------------------------------------------------
-- Algebraic properties

+-cong : Congruent₂ _+_
+-cong x⊒y u⊒v as p with split as p
... | inj₁ p′ = join₁ as (x⊒y as p′)
... | inj₂ p′ = join₂ as (u⊒v as p′)

+-assoc : Associative _+_
+-assoc r₁ r₂ r₃ [] (+ₗ p) = +ₗ (+ₗ p)
+-assoc r₁ r₂ r₃ [] (+ᵣ (+ₗ p)) = +ₗ (+ᵣ p)
+-assoc r₁ r₂ r₃ [] (+ᵣ (+ᵣ p)) = +ᵣ p
+-assoc r₁ r₂ r₃ (a ∷ as) p = +-assoc _ _ _ as p

+-comm : Commutative _+_
+-comm r₁ r₂ [] (+ₗ p) = +ᵣ p
+-comm r₁ r₂ [] (+ᵣ p) = +ₗ p
+-comm r₁ r₂ (a ∷ as) p = +-comm _ _ as p

+-idem : Idempotent _+_
+-idem r [] p = +ₗ p -- annoyingly not unique
+-idem r (a ∷ as) p = +-idem _ as p

+-identityˡ : LeftIdentity Λ _+_
+-identityˡ r [] p = +ᵣ p
+-identityˡ r (a ∷ as) p = +-identityˡ _ as p

+-identityʳ : RightIdentity Λ _+_
+-identityʳ r [] p = +ₗ p
+-identityʳ r (a ∷ as) p = +-identityʳ _ as p

⨟-cong : Congruent₂ _⨟_
⨟-cong x⊒y u⊒v [] (p₁ ⨟ p₂) = x⊒y [] p₁ ⨟ u⊒v [] p₂
⨟-cong {x = x} {y = y} x⊒y u⊒v (a ∷ as) p with A? x | A? y
... | no ¬Ax | no ¬Ay = ⨟-cong (λ as′ → x⊒y (a ∷ as′)) u⊒v as p
... | no ¬Ax | yes Ay = contradiction (x⊒y [] Ay) ¬Ax
... | yes Ax | no ¬Ay = join₂ as (⨟-cong (λ as′ → x⊒y (a ∷ as′)) u⊒v as p)
... | yes Ax | yes Ay with split as p
... | inj₁ p′ = join₁ as (u⊒v (a ∷ as) p′)
... | inj₂ p′ = join₂ as (⨟-cong (λ as′ → x⊒y (a ∷ as′)) u⊒v as p′)

⨟-distribˡ-+ : _⨟_ DistributesOverˡ _+_
⨟-distribˡ-+ r₁ r₂ r₃ [] (+ₗ (p₁ ⨟ p₂)) = p₁ ⨟ +ₗ p₂
⨟-distribˡ-+ r₁ r₂ r₃ [] (+ᵣ (p₁ ⨟ p₂)) = p₁ ⨟ +ᵣ p₂
⨟-distribˡ-+ r₁ r₂ r₃ (a ∷ as) p with A? r₁
... | no ¬Ar₁ with split as p
... | inj₁ p′ = ⨟-distribˡ-+ _ _ _ as (join₁ as p′)
... | inj₂ p′ = ⨟-distribˡ-+ _ _ _ as (join₂ as p′)
⨟-distribˡ-+ r₁ r₂ r₃ (a ∷ as) p | yes Ar₁ with split as p
... | inj₁ p′ with split as p′
... | inj₁ p″ = join₁ as (join₁ as p″)
... | inj₂ p″ = join₂ as (⨟-distribˡ-+ _ _ _ as (join₁ as p″))
⨟-distribˡ-+ r₁ r₂ r₃ (a ∷ as) p | yes Ar₁ | inj₂ p′ with split as p′
... | inj₁ p″ = join₁ as (join₂ as p″)
... | inj₂ p″ = join₂ as (⨟-distribˡ-+ _ _ _ as (join₂ as p″))

⨟-distribʳ-+ : _⨟_ DistributesOverʳ _+_
⨟-distribʳ-+ r₁ r₂ r₃ [] (+ₗ (p₁ ⨟ p₂)) = +ₗ p₁ ⨟ p₂
⨟-distribʳ-+ r₁ r₂ r₃ [] (+ᵣ (p₁ ⨟ p₂)) = +ᵣ p₁ ⨟ p₂
⨟-distribʳ-+ r₁ r₂ r₃ (a ∷ as) p with A? r₂ | A? r₃
... | no ¬Ar₂ | no ¬Ar₃ = ⨟-distribʳ-+ _ _ _ as p
... | no ¬Ar₂ | yes Ar₃ with split as p
... | inj₁ p′ = join₂ as (⨟-distribʳ-+ _ _ _ as (join₁ as p′))
... | inj₂ p′ with split as p′
... | inj₁ p″ = join₁ as p″
... | inj₂ p″ = join₂ as (⨟-distribʳ-+ _ _ _ as (join₂ as p″))
⨟-distribʳ-+ r₁ r₂ r₃ (a ∷ as) p | yes Ar₂ | no ¬Ar₃ with split as p
... | inj₂ p′ = join₂ as (⨟-distribʳ-+ _ _ _ as (join₂ as p′))
... | inj₁ p′ with split as p′
... | inj₁ p″ = join₁ as p″
... | inj₂ p″ = join₂ as (⨟-distribʳ-+ _ _ _ as (join₁ as p″))
⨟-distribʳ-+ r₁ r₂ r₃ (a ∷ as) p | yes Ar₂ | yes Ar₃ with split as p
... | inj₁ p′ with split as p′
... | inj₁ p″ = join₁ as p″
... | inj₂ p″ = join₂ as (⨟-distribʳ-+ _ _ _ as (join₁ as p″))
⨟-distribʳ-+ r₁ r₂ r₃ (a ∷ as) p | yes Ar₂ | yes Ar₃ | inj₂ p′ with split as p′
... | inj₁ p″ = join₁ as p″
... | inj₂ p″ = join₂ as (⨟-distribʳ-+ _ _ _ as (join₂ as p″))

⨟-assoc : Associative _⨟_
⨟-assoc r₁ r₂ r₃ [] (p₁ ⨟ (p₂ ⨟ p₃)) = (p₁ ⨟ p₂) ⨟ p₃
⨟-assoc r₁ r₂ r₃ (a ∷ as) p with A? r₁ | A? r₂
... | no ¬Ar₁ | no ¬Ar₂ = ⨟-assoc _ _ _ as p
... | no ¬Ar₁ | yes Ar₂ = ⨟-assoc _ _ _ as p
... | yes Ar₁ | no ¬Ar₂ with split as p
... | inj₁ p′ = ⨟-distribʳ-+ _ _ _ as (join₁ as p′)
... | inj₂ p′ = ⨟-distribʳ-+ _ _ _ as (join₂ as (⨟-assoc _ _ _ as p′))
⨟-assoc r₁ r₂ r₃ (a ∷ as) p | yes Ar₁ | yes Ar₂ with split as p
... | inj₂ p′ = join₂ as (⨟-distribʳ-+ _ _ _ as (join₂ as (⨟-assoc _ _ _ as p′)))
... | inj₁ p′ with split as p′
... | inj₁ p″ = join₁ as p″
... | inj₂ p″ = join₂ as (⨟-distribʳ-+ _ _ _ as (join₁ as p″))

⨟-zeroˡ : LeftZero Λ _⨟_
⨟-zeroˡ r as p = contradiction p (end as)

⨟-zeroʳ : RightZero Λ _⨟_
⨟-zeroʳ r as p = contradiction p (end as)

⨟-identityˡ : LeftIdentity ε _⨟_
⨟-identityˡ r [] p = star ⨟ p
⨟-identityˡ r (a ∷ as) p = join₁ as p

⨟-identityʳ : RightIdentity ε _⨟_
⨟-identityʳ r [] p = p ⨟ star
⨟-identityʳ r (a ∷ as) p with A? r
... | no ¬Ar = ⨟-identityʳ _ as p
... | yes Ar = join₂ as (⨟-identityʳ _ as p)
